/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  |           =====>>>>> 2002 <<<<=====             |
  |                                                 |
  \-------------------------------------------------/
  
  ===================================================
  
  Name:       ccm.C
  
  @Copyright: Damian J. J. Farnell
  
  Date:       Wed Sep 25 15:32:06 BST 2002
  
  Purpose:    Determine and solve CCM equations.
  
  ===================================================
  
  ---------------------------------------------------

  MODULE NAME: CCM.C 

  PURPOSE:     MAIN ROUTINE WHICH CALLS EACH MODULE

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

// #include "nr.h"
// #include "nrutil.h"
#include "Constants.h"
#include "GlobalVariables.h"
#include "FunctionPrototypes.h"
#include "version.h"

#if PARALLEL_FLAG
#include <mpi.h>
#else
#define MPI_Barrier(xxx)    // dummy code if not parallel
#endif

// Run in parallel with:  mpirun -n 1 -p ./ccm.x 

void obtainExcitations(void);

int main(int argc, char *argv[])
{
  
  // int i;
  int run_option=7;  /* default = bit0..2 set = 1+2+4 */
  int  t2;   /* measure speed */

#if (PARALLEL_FLAG)

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD,&npe);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);

  if (!myid)  // reduce output
    printf("NPE = %d MYID = %d\n", npe, myid);

#else

  npe=1;
  myid=0;

#endif

  /* ccm.C: 

     Purpose: To enumerate all CCM spin cluster configurations within 
     an LSUBm (or other) approach and then determine and solve the
     CCM ground ket- and bra-state equations.

     ====> The Code Is Split Into 5 Modules: <====

     1) Input.C

     ..(a) Read In Control Variables and Initialise Data.

     ..(b) Define the crystallographic lattice.

     2) Define.C

     ..(a) Enumerate the connected clusters.

     ..(b) Determine the unconnected clsuters.

     ..(c) Choose a subset of these clusters (e.g., those
     ......which conserve s_T^z).

     3) Obtain.C

     ..(a) Loop over total number of fundamental configs.
     
     ..(b) Determine the CCM equations for each `term.'
  
     ..(c) Save CCM equations to disk.

     4. Collect.C

     ..(a) Collects like terms together and thus reduces disk usage.

     5. Solve.C

     ..(a) Solve ket- and bra-state equations.
     ..(b) Find ground-state energy, sublattice magnetisation, etc.

  */

  seconds();  /* initialize unix timer */

  // this is info for debugging and better user feedback
  if (!myid) printf("CCM for Quantum Spin Systems, version " Version "\n");
  if (!myid) printf("int=%d long=%d size_t=%d\n",
             8*(int)sizeof(int),    8*(int)sizeof(long),
             8*(int)sizeof(size_t));
#ifdef _UNISTD_H
  if (!myid) printf("off_t=%d\n", 8*(int)sizeof(off_t));
#endif
  if (!myid) {
#ifdef __GNUC__
    printf("Compiler: GNUC %d.%d\n",__GNUC__,__GNUC_MINOR__);
#endif
#ifdef __DECCXX_VER
    printf("Compiler: DECCXX %d\n",__DECCXX_VER);
#endif
#ifdef __SUNPRO_CC
    printf("Compiler: SUNPRO_CC 0x%x\n",__SUNPRO_CC);
#endif
#if defined(__mips) && defined(_COMPILER_VERSION)
    printf("Compiler: MIPS_CC %d\n",_COMPILER_VERSION);
#endif
  }

  Control.excitations=0;
  if (argc >= 3) {  // allow additional options
    int i;
    run_option=0;               // we convert run_option to bitpattern
    for (i=0;argv[argc-1][i];i++) {  // "23" -> 6, "123" = "4" -> 7, "1" -> 1
      switch(argv[argc-1][i]) {      // bit0 for part1, bit1 for part2, ...
        case '1': run_option|=1; break;  // compute ccm-configs
        case '2': run_option|=2; break;  // compute ccm-equations
        case '3': run_option|=4; break;  // solve equations
        case '4': run_option|=7; break;  // all together
        case 'x': Control.excitations=1; break;  // compute excitations 
        default: printf("Warning: invalid run_option ignored\n");
      }
    }
    if (!myid)
      printf("Run option is =%s converted to %d\n", argv[argc-1], run_option);
  } else {
    printf("Arguments: input_script_name.ccm run_option_1234x\n"
           " 123 - same as old option 4, do all essential steps topgether\n" 
           "   x - compute excited states\n"
           "Exiting!!\n");
    ERR(1);
  }

  /* dont buffer output to get all outputs immediately */
  fflush(stdout); setvbuf(stdout,NULL,_IONBF,0);

  /* be sure all threads have runoption set */
  MPI_Barrier(MPI_COMM_WORLD);

  // ToDo: better set Control.name here

  input(run_option, argv[argc-2]); // read and distribute .configs at end

  MPI_Barrier(MPI_COMM_WORLD);

  if (run_option & 1) define();   // compute and write .configs

  t2 = seconds();        /* get unix time in seconds */
  MPI_Barrier(MPI_COMM_WORLD);
  if (!myid)
    printf(" define finished after t[min]= %6.2f\n", t2/60.);

  MPI_Barrier(MPI_COMM_WORLD);

#ifndef USE_FILE
  if ((run_option&6)==2 && myid==0)
     { printf("\nWarning: Terms will not be saved! try USE_FILE!\n"); }
  if ((run_option&6)==4 && myid==0)
     { printf("\nError: No term file available! try USE_FILE!\n"); exit(1); }
#else
  if ((run_option&6)==6 && myid==0)
     { printf("\nWarning: undef USE_FILE may be much faster\n"); }
#endif

  if (run_option & 2) {
    obtain();  /* get and store collected CCM terms to data.terms */
    /* ToDo: obtain is badly balanced 
     *    we should start with slowest term and put the next easier term to
     *    the next free mpi-jobs until all jobs are done
     *    output 3 biggest and smalles number of terms per cluster? 
     */
    /* May2008 DJJF: see obtainGeneralisedExpectations.C */
    if (Control.greatExpectations) obtainGenExp();
    /* Jun2008 DJJF: see obtainExcitatios.C */
    if (Control.excitations) obtainExcitations();
  }
  /* "OBTAIN ... SUCCESSFULLY COMPLETED" outputted */
  t2=seconds();        /* get unix time in seconds */
  MPI_Barrier(MPI_COMM_WORLD);  /* 100% CPU, nice 19 */
  if (!myid)
    printf(" obtain finished after t[min]= %6.2f\n", t2/60.);
  MPI_Barrier(MPI_COMM_WORLD); /* wait for other outputs */
  

  if (run_option & 4) 
    solve();   /* read data.terms (often) and output data.output */

  t2=seconds();        /* get unix time in seconds */
  if (!myid)
    printf(" solve finished after t[min]= %6.2f\n\n", t2/60.);

  if (!myid)
    printf("We'd be grateful if you would acknowledge\n"
           "the use of this CCCM package in your publication.\n");

  // if (myid==0)
  //  printf("\nPress a key to close this window.\n\n");
  //  getchar();

#if (PARALLEL_FLAG)
  MPI_Finalize();
#endif
  return 0;
}
